#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <netdb.h>
#include <errno.h>
#include <err.h>

int errno;

typedef struct {
	int		 status;
	int	     socket;
	int      port;
	int      lock;
	int      maxclients;
	int      nclients;
} srv_t, *psrv_t;

#define _srv_c_
#include "srv.h"
#undef _srv_c_

void
srv_lock(psrv_t s) {
	if (!s) return;
	while (s->lock); 
	s->lock = 1;
}

void 
srv_unlock(psrv_t s) {
	if (!s) return;
	if (s->lock) s->lock = 0;
}

psrv_t 
srv_destroy(psrv_t s) {
	if (s) {
		s->status = srvUNKNOWN;
		close(s->socket);
		free(s);
	}
	return NULL;
}

size_t srv_sizeof() { return sizeof(srv_t); }

psrv_t
srv_new(int port, int maxclients) {
	static long on = 1L;
 	struct sockaddr_in sin;             
	psrv_t s;

	if (!(s = (psrv_t) malloc(sizeof(srv_t)))) {
		err_error("cannot allocate memory for srv\n"); 
		return NULL;
	}

	memset((void *) s, 0, sizeof(srv_t));

	s->maxclients = maxclients;
	s->port       = port;
	s->status = srvCREATING;	

	if ((s->socket = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		err_error("cannot create socket : %s\n", strerror(errno));
		return srv_destroy(s);
	}

	s->status = srvCONFIGURING;

 	setsockopt(s->socket, SOL_SOCKET, SO_REUSEADDR, (void *) &on, sizeof(int));
 	setsockopt(s->socket, SOL_SOCKET, SO_KEEPALIVE, (void *) &on, sizeof(int));

	s->status = srvBINDING;

    bzero(&sin, sizeof(struct sockaddr_in));
    sin.sin_family = AF_INET;
    sin.sin_port = htons(s->port);
    sin.sin_addr.s_addr = htonl(INADDR_ANY); 
    /* sin.sin_addr.s_addr = INADDR_ANY; */

	if (bind(s->socket, (struct sockaddr *) &sin, sizeof(sin))) {
		err_error("cannot bind socket to port : %s\n", strerror(errno));
		return srv_destroy(s);
	}
		
	s->status = srvCONNECTING;

	if (listen(s->socket, maxclients)) {
		err_error("cannot listen on port : %s\n", strerror(errno));
		return srv_destroy(s);
	}

	s->status = srvLISTENING;
	return s;
}

int
srv_port(psrv_t s) {
	if (!s) return 0;
	return s->port;
}

int srv_socket(psrv_t s) {
	if (!s) return 0;
	return s->socket;
}

int srv_maxclients(psrv_t s) {
	if (!s) return 0;
	return s->maxclients;
}

int srv_status(psrv_t s) {
	if (!s) return 0;
	return s->status;
}

int
srv_clients_connected(psrv_t s) {
	if (!s) return -1;
	return s->nclients;
}

int
srv_client_connect(psrv_t s) {
	int fd;
	struct sockaddr	cli_addr;
	size_t len;

	srv_lock(s);

#if 0
	if (s->nclients >= s->maxclients) {
		err_error("srv no more client connexion\n");
		return -1;
	}
#endif

	if ((fd = accept(s->socket, &cli_addr, &len)) < 0) {
		err_error("cannot accept connections : %s\n", strerror(errno));
		return -2;
	}
	s->nclients++;

	/* printf("%s is connected\n", cli_addr.h_name); */
	
	srv_unlock(s);
	return fd;	
}

int
srv_client_disconnect(psrv_t s, int cid) {
	if (!s) return -1;
	srv_lock(s);	
	s->nclients--;
	close(cid);
	srv_unlock(s);
	return 0;
}

int
srv_recv_to(int cid, char *data, size_t size, int sec, int usec) {
	int r;
	struct timeval time;
	fd_set fdread;

	time.tv_sec  = sec;
	time.tv_usec = usec;

	FD_SET(cid, &fdread);
	
	r = select(cid + 1, &fdread, (fd_set *) 0, (fd_set *) 0, &time);
		
	if (r == 0) {
		err_error("timeout occured aftrer %d.%d seconds\n", sec, usec);
		return 0;
	}
	if (r < 0) {
		err_error("select error : %s\n", strerror(errno));
		return 0;
	}
		
	return read(cid, data, size);
}

int
srv_recv(int cid, char *data, size_t size) {
	if (!data) return 0;
	return read(cid, (void *) data, size);
}

int
srv_send(int cid, char *data, size_t size) {
	return write(cid, (void *) data, size);
}

